from cpython cimport PyObject
from numpy cimport dtype as npdtype, npy_int64, npy_intp

from medvedi.native.cpython cimport PyTypeObject


cdef extern from "numpy/arrayobject.h":
    PyTypeObject PyArray_Type
    PyTypeObject PyDatetimeArrType_Type
    PyTypeObject PyDoubleArrType_Type
    PyTypeObject PyIntegerArrType_Type
    PyTypeObject PyFloatArrType_Type
    PyTypeObject PyTimedeltaArrType_Type

    enum: NPY_DATETIME_NAT

    ctypedef struct PyArray_Descr:
        char kind
        char type
        char byteorder
        char flags
        int type_num
        int itemsize "elsize"
        int alignment

    PyObject *PyArray_NewFromDescr(
        PyTypeObject *subtype,
        PyArray_Descr *descr,
        int nd,
        const npy_intp *dims,
        const npy_intp *strides,
        void *data,
        int flags,
        PyObject *obj,
    )
    npdtype PyArray_DescrNew(npdtype)

    void *PyArray_DATA(PyObject *) nogil
    char *PyArray_BYTES(PyObject *) nogil
    npy_intp PyArray_DIM(PyObject *, size_t) nogil
    npy_intp PyArray_STRIDE(PyObject *, size_t) nogil
    int PyArray_NDIM(PyObject *) nogil
    npy_intp PyArray_ITEMSIZE(PyObject *) nogil
    bint PyArray_CheckExact(PyObject *) nogil
    PyArray_Descr *PyArray_DESCR(PyObject *) nogil
    int PyArray_TYPE(PyObject *) nogil
    bint PyArray_IS_C_CONTIGUOUS(PyObject *) nogil
    bint PyArray_IS_F_CONTIGUOUS(PyObject *) nogil
    void PyArray_ScalarAsCtype(PyObject *scalar, void *ctypeptr) nogil

    ctypedef enum NPY_DATETIMEUNIT:
        NPY_FR_ERROR = -1
        NPY_FR_M = 1
        NPY_FR_W = 2
        NPY_FR_D = 4
        NPY_FR_h = 5
        NPY_FR_m = 6
        NPY_FR_s = 7
        NPY_FR_ms = 8
        NPY_FR_us = 9
        NPY_FR_ns = 10
        NPY_FR_ps = 11
        NPY_FR_fs = 12
        NPY_FR_as = 13
        NPY_FR_GENERIC = 14

    ctypedef struct PyArray_DatetimeMetaData:
        NPY_DATETIMEUNIT base
        int num

    ctypedef struct PyDatetimeScalarObject:
        npy_int64 obval
        PyArray_DatetimeMetaData obmeta



# cython: language_level=3, boundscheck=False, nonecheck=False, optimize.unpack_method_calls=True
# cython: warn.maybe_uninitialized=True

from cpython.pycapsule cimport PyCapsule_GetPointer
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.unordered_map cimport pair, unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.vector cimport vector


cdef extern from "mi_heap_destroy_stl_allocator.h" nogil:
    cdef cppclass mi_heap_destroy_stl_allocator[T]:
        mi_heap_destroy_stl_allocator() except +
        mi_heap_destroy_stl_allocator(const mi_heap_destroy_stl_allocator &)
        T* allocate(size_t count) except +
        void deallocate(T*)

    cdef cppclass mi_unordered_map[T, U, HASH=*, PRED=*](unordered_map[T, U, HASH, PRED]):
        mi_unordered_map mi_unordered_map[X](mi_heap_destroy_stl_allocator[X]&) except +
        pair[mi_unordered_map.iterator, bool] try_emplace(...) except +
        mi_heap_destroy_stl_allocator[T] get_allocator()

    cdef cppclass mi_unordered_set[T, HASH=*, PRED=*](unordered_set[T, HASH, PRED]):
        mi_unordered_set mi_unordered_set[X](mi_heap_destroy_stl_allocator[X]&) except +
        pair[mi_unordered_set.iterator, bool] emplace(...) except +
        mi_heap_destroy_stl_allocator[T] get_allocator()

        mi_unordered_set.iterator erase(mi_unordered_set.iterator)
        mi_unordered_set.iterator erase(mi_unordered_set.iterator, mi_unordered_set.iterator)
        size_t erase(T&)

    cdef cppclass mi_vector[T](vector[T]):
        mi_vector mi_vector[X](mi_heap_destroy_stl_allocator[X]&) except +
        T& emplace_back(...) except +
        mi_heap_destroy_stl_allocator[T] get_allocator()

    cdef cppclass mi_string(string):
        mi_string mi_string[X](const char *, size_t, mi_heap_destroy_stl_allocator[X]&) except +
        mi_heap_destroy_stl_allocator[char] get_allocator()


cdef inline mi_heap_destroy_stl_allocator[char] *mi_heap_allocator_from_capsule(obj) except? NULL:
    return <mi_heap_destroy_stl_allocator[char] *> PyCapsule_GetPointer(obj, b"mi_heap_destroy_stl_allocator")


cdef inline void _delete_mi_heap_allocator_in_capsule(obj):
    cdef mi_heap_destroy_stl_allocator[char] *alloc = mi_heap_allocator_from_capsule(obj)
    del alloc
